import java.io.*;
import java.util.*;

public class CalcParser {
    private PushbackReader reader;
    private Token token = null;

    public class Token {
	public String type;
	public String str;

	public String toString() {
	    return str;
	}
    }

    public class Expression {
	public Token op = null;
	public Expression[] args = null;

	public Expression(Token op) {
	    this.op = op;
	    args = null;
	}

	public Expression(Token op, Expression arg1) {
	    this.op = op;
	    args = new Expression[1];
	    args[0] = arg1;
	}

	public Expression(Token op, Expression arg1, Expression arg2) {
	    this.op = op;
	    args = new Expression[2];
	    args[0] = arg1;
	    args[1] = arg2;
	}

	public Expression(Token op, List args) {
	    this.op = op;
	    this.args = null;
	    if (args != null && args.size() > 0) {
		this.args = new Expression[1];
		this.args = (Expression[])args.toArray(this.args);
	    }
	}

	public Expression(Token op, Expression[] args) {
	    this.op = op;
	    this.args = args;
	}

	public int argc() {
	    return args == null ? 0 : args.length;
	}

	public boolean isInteger() {
	    return op.type.equals("INTEGER");
	}

	public boolean isFunctor(String name) {
	    return op.type.equals("SYMBOL") && op.str.equals(name);
	}

	public boolean isFunctor(String name, int arity) {
	    return isFunctor(name) && argc() == arity;
	}

	public boolean isSymbol(String name) {
	    return isFunctor(name, 0);
	}

	public boolean isSymbol() {
	    return op.type.equals("SYMBOL") && argc() == 0;
	}

	public boolean isOperator(String name) {
	    return op.type.equals(name);
	}

	public boolean isOperator(String name, int arity) {
	    return isOperator(name) && argc() == arity;
	}

	public String toString() {
	    String s = "";
	    if (args == null) {
		s = op.toString();
	    } else {
		String delim = "";
		for (int i = 0; i < args.length; i++) {
		    s += delim + args[i].toString();
		    delim = ";";
		}
		s = op.toString() + "(" + s + ")";
	    }
	    return s;
	}

    }

    public CalcParser(String inputString) {
	this(new PushbackReader(new StringReader(inputString)));
    }

    public CalcParser(PushbackReader reader) {
	this.reader = reader;
	token = null;
    }

    private boolean isWhitespace(int c) {
	return Character.isWhitespace((char)c);
    }

    private boolean isDigit(int c) {
	return Character.isDigit((char)c);
    }

    private boolean isSymbolStart(int c) {
	return (Character.isUnicodeIdentifierStart((char)c) || c == '$')
	    && ! Character.isIdentifierIgnorable((char)c);
    }

    private boolean isSymbolPart(int c) {
	return (Character.isUnicodeIdentifierPart((char)c) || c == '$' || c == ':' || c == '.')
	    && ! Character.isIdentifierIgnorable((char)c);
    }

    private void unread(int c) throws IOException {
	if (c >= 0) {
	    reader.unread(c);
	}
    }

    public Token nextToken() {
	Token token = null;
	StringBuffer type = new StringBuffer();
	StringBuffer str = new StringBuffer();
	try {
	    int c = reader.read();
	    if (c < 0) {
		return null;
	    } else if (isWhitespace(c)) {
		type = new StringBuffer(" ");
		while  (isWhitespace(c)) {
		    c = reader.read();
		}
		unread(c);
	    } else if (isDigit(c)) {
		type = new StringBuffer("INTEGER");
		do {
		    str.append((char)c);
		    c = reader.read();
		} while (isDigit(c));
		unread(c);
	    } else if (c == '"') {
		type = new StringBuffer("STRING");
		while (true) {
		    c = reader.read();
		    if (c == '"') {
			c = reader.read();
			if (c != '"') {
			    unread(c);
			    break;
			}
		    }
		    str.append((char)c);
		}
	    } else if (isSymbolStart(c)) {
		type = new StringBuffer("SYMBOL");
		do {
		    str.append(Character.toUpperCase((char)c));
		    c = reader.read();
		} while (isSymbolPart(c));
		unread(c);
	    } else if (c == '<') {
		type.append((char)c);
		str.append((char)c);
		c = reader.read();
		if (c == '=' || c == '>') {
		    type.append((char)c);
		    str.append((char)c);
		} else {
		    unread(c);
		}
	    } else if (c == '>') {
		type.append((char)c);
		str.append((char)c);
		c = reader.read();
		if (c == '=') {
		    type.append((char)c);
		    str.append((char)c);
		} else {
		    unread(c);
		}
	    } else {
		type.append((char)c);
		str.append((char)c);
	    }
	    token = new Token();
	    token.type = type.toString();
	    token.str = str.toString();
	} catch (IOException e) {
	}
	return token;
    }

    private void skipSpaces() {
	while (token != null && token.type.equals(" ")) {
	    token = nextToken();
	}
    }

    // Expression ::= Expr { CompOp Expr }
    // CompOp ::= '=' | '<>' | '<' | '<=' | '>' | '>='
    private Expression parseExpression() throws IllegalArgumentException {
	Expression x, y;
	Token op;
	x = parseExpr();
	for (skipSpaces(); token != null; skipSpaces()) {
	    if (token.type.equals("=") || token.type.equals("<>")
		|| token.type.equals("<") || token.type.equals("<=")
		|| token.type.equals(">") || token.type.equals(">=")) {
		op = token;
		token = nextToken();
		y = parseExpr();
		x = new Expression(op, x, y);
	    } else {
		break;
	    }
	}
	return x;
    }

    // Expr ::= [ '+' | '-' ] Term { ( '+' | '-' ) Term }
    private Expression parseExpr() throws IllegalArgumentException {
	Expression x, y;
	Token op;
	skipSpaces();
	if (token == null) {
	    return null;
	} else if (token.type.equals("+") || token.type.equals("-")) {
	    op = token;
	    token = nextToken();
	    y = parseTerm();
	    x = new Expression(op, y);
	} else {
	    x = parseTerm();
	}
	for (skipSpaces(); token != null; skipSpaces()) {
	    if (token.type.equals("+") || token.type.equals("-")) {
		op = token;
		token = nextToken();
		y = parseTerm();
		x = new Expression(op, x, y);
	    } else {
		break;
	    }
	}
	return x;
    }

    // Term ::= Factor { ( '*' | '/' ) Factor }
    private Expression parseTerm() throws IllegalArgumentException {
	Expression x, y;
	Token op;
	x = parseFactor();
	for (skipSpaces(); token != null; skipSpaces()) {
	    if (token.type.equals("*") || token.type.equals("/")) {
		op = token;
		token = nextToken();
		y = parseFactor();
		x = new Expression(op, x, y);
	    } else {
		break;
	    }
	}
	return x;
    }

    // Factor ::= Integer | String | Symbol | Symbol '(' Expression { ';' Expression } ')'
    //          | '(' Expression ')'
    private Expression parseFactor() throws IllegalArgumentException {
	Expression x, y;
	Token op;
	skipSpaces();
	if (token == null) {
	    return null;
	} else if (token.type.equals("INTEGER") || token.type.equals("STRING")) {
	    op = token;
	    token = nextToken();
	    x = new Expression(op);
	} else if (token.type.equals("SYMBOL")) {
	    op = token;
	    token = nextToken();
	    skipSpaces();
	    if (token != null && token.type.equals("(")) {
		ArrayList args = new ArrayList();
		do {
		    token = nextToken();
		    y = parseExpression();
		    args.add(y);
		    skipSpaces();
		} while (token != null && token.type.equals(";"));
		if (token == null || ! token.type.equals(")")) {
		    throw new IllegalArgumentException(") is expected at " + token);
		}
		token = nextToken();
		x = new Expression(op, args);
/*
	    } else if (token != null && token.type.equals(":")) {
		x = new Expression(op);
		op = token;
		token = nextToken();
		y =  parseFactor();
		x = new Expression(op, x, y);
*/
	    } else {
		x = new Expression(op);
	    }
	} else if (token.type.equals("(")) {
	    token = nextToken();
	    x = parseExpression();
	    skipSpaces();
	    if (token == null || ! token.type.equals(")")) {
		throw new IllegalArgumentException(") is expected at " + token);
	    }
	    token = nextToken();
	} else {
	    throw new IllegalArgumentException("can not start expression at " + token);
	}
	return x;
    }

    public Expression parse() throws IllegalArgumentException {
	token = nextToken();
	skipSpaces();
	if (token.type.equals("=")) {
	    token = nextToken();
	}
	Expression x = parseExpression();
	skipSpaces();
	if (token != null) {
	    throw new IllegalArgumentException("unknown operator at " + token);
	}
	return x;
    }
    
    public static void main(String[] args) {
	CalcParser parser = new CalcParser(args[0]);
	try {
	    Expression expr = parser.parse();
	    System.out.println(expr);
	} catch (IllegalArgumentException e) {
	    System.out.println(e);
	}
    }

}
